from multiprocessing import Process, Queue, Manager
import multiprocessing  
import zmq
from functools import wraps
import logging 
import datetime
import json 
import time 
import traceback 
import copy
import threading
import queue 
import redis 
import json 
import os 
from dateutil import parser

from ..sparkLib import SparkLib,OrderSocket
from ..sparkLib.utility import dataws

BROKERS = {
    "IIFLXTSDATA" : dataws.IIFLXTSWS,
    "KOTAKNEO" : dataws.Kotakws,
    "ANGELONE" : dataws.AngelWs,
    "FYERS" : dataws.FyersWs,
    "SHOONYA" : dataws.ShoonyaWS,
    "UPSTOX" : dataws.UpstoxWS
}

BROKERconf = {
    "IIFLXTSDATA" : {"maxConnection" : 1, "tokenLimit" : 800, "OI" : False, 
                     "is_depth" : True, "segments" : ['NSE', 'BSE', "NFO", "BFO"]},
    "ANGELONE" : {"maxConnection" : 3, "tokenLimit":1000, "OI" : True,
               "is_depth" : True, "segments" : ['NSE', "BSE", "NFO", "BFO", "MCX"]},
    "KOTAKNEO" : {"maxConnection" : 1, "tokenLimit":3000, "OI" : True,
               "is_depth" : True, "segments" : ['NSE', "BSE", "NFO", "BFO", "MCX"]},
    "SHOONYA" : {"maxConnection" : 1, "tokenLimit" : 3000, "OI" : True,
                 "is_depth" : True, "segments" : ['NSE', "BSE", "NFO", "BFO", "MCX"]},
    "FYERS" : {"maxConnection" : 1, "tokenLimit": 5000, "OI" : False,
               "is_depth" : True, "segments" : ['NSE', "BSE", "NFO", "BFO", "MCX"]},
    "UPSTOX" : {"maxConnection" : 1, "tokenLimit": 100, "OI" : True,
               "is_depth" : True, "segments" : ['NSE', "BSE", "NFO", "BFO", "MCX"]},
}

class OrderWS():
    def __init__(self, sessionid, ordPubSub, logger) : 
        self.sessionid = sessionid
        self.ordPubSub = ordPubSub
        self.logger = logger

    def _zmq_config(self, host, port):
        context = zmq.Context()
        socket = context.socket(zmq.PUB)
        socket.bind("tcp://{host}:{port}".format(host=host, port=str(port)))
        return socket
    
    def ordersListner(self, message):
        self.pubsocket.send_string("{message}".format(message=message))

    def on_error(self, error):
        self.logger.critical(f"FROM ORDERWS : Error - {str(error)}")

    def on_close(self):
        self.logger.critical(f"FROM ORDERWS : Connection Closed.")

    def on_open(self):
        self.logger.critical(f"FROM ORDERWS : Connection Established.")

    def close(self):
        self.order_socket.close_connection()
    
    def disconnect(self):
        self.order_socket.close_connection()

    def connect(self):
        self.pubsocket = self._zmq_config(**self.ordPubSub)
        self.order_socket = OrderSocket(
            access_token=self.sessionid,
            on_order=self.ordersListner,
            on_error=self.on_error,
            on_close=self.on_close,
            on_connect=self.on_open,
            reconnect=True,             # Optional: Attempt to reconnect if connection is closed
            max_reconnect_attempts=20   # Optional: Maximum attempts to reconnect
        )
        self.order_socket.connect()

class brokerWS():
    def __init__(self, accountData, sparkSession, respQueue, subsQueue, connectionType = "ltp", spwID = None) :
        self.accountData = accountData
        self.identifier = ":".join([self.accountData['UCC'], self.accountData['AccountName']])
        self._function = BROKERS[self.accountData['Broker']]
        self.broker = self.accountData['Broker']
        self.respQueue = respQueue
        self.connectionType = connectionType
        self.subsQueue = subsQueue
        self.run = True
        self.sparkSession = sparkSession
        self.spwID = spwID
        self.maxDelay = 3

    def runAction(self):
        while self.run :
            data = self.subsQueue.get()
            if data['action'] == "subscribe" : 
                self.socket.Subscribe(data['tokens'])
            
            elif data['action'] == "unsubscribe" : 
                self.socket.Unsubscribe(data['tokens'])
            time.sleep(0.1)

    def onOpen(self, *args):
        resp = {"type" : "GENERAL", "Identifier" : self.identifier, "data" : [str(list(args))], "message" : f"Connection Opened - {self.identifier}", "Broker" : self.broker, 
                "spwID" : self.spwID}
        self.respQueue.put(resp)

    def onClose(self, *args, **kwargs):
        resp = {"type" : "CLOSE", "Identifier" : self.identifier, "data" : [str(list(args))], "message" : f"Connection Closed - {self.identifier}", "Broker" : self.broker,
                "spwID" : self.spwID}
        self.respQueue.put(resp)

    def onLtp(self, data):
        resp = {"type" : "DATA", "Identifier" : self.identifier, "data" : data, "message" : "", "Broker" : self.broker, "spwID" : self.spwID}
        if parser.parse(data['timestamp_str']) + datetime.timedelta(seconds=self.maxDelay) < datetime.datetime.now() :
            # print(resp['Broker'], " DELAY")
            # print(parser.parse(data['timestamp_str']) + datetime.timedelta(seconds=self.maxDelay), datetime.datetime.now())
            pass
        else: 
            self.respQueue.put(resp)

    def onError(self, data, *args, **kwargs):
        resp = {"type" : "ERROR" , "Identifier" : self.identifier, "data" : data, "message" : f"Error - {self.identifier}", "Broker" : self.broker, 
                "spwID" : self.spwID}
        self.respQueue.put(resp)

    def onDelay(self):
        pass
    
    def onDepth(self, data):
        pass
    
    def connect(self):
        self.socket = self._function(accountData=self.accountData,
                       accessToken = self.sparkSession,
                       dataType=self.connectionType,
                       onLtp= self.onLtp, 
                       onDepth=self.onDepth,
                       onError=self.onError,
                       onClose=self.onClose,
                       onOpen=self.onOpen)
        
        if self.broker == "UPSTOX" : 
            threading.Thread(target=self.runAction).start()
            self.socket.connect()
        else: 
            threading.Thread(target=self.socket.connect).start()
            threading.Thread(target=self.runAction).start()

class TickHandler():
    tickQueue = queue.Queue()
    MINdata = {}
    startTime = datetime.time(9,15)
    endTime = datetime.time(15,30)
    premindata = {}
    prev_minval = 0
    minvalList = []
    __delayTrack = {}
    __maxDelay = 3
    
    def __init__(self, ltpPubSub, respQueue, obj, logger, redisConf, _toRedis = False, _priorityTicks = False):
        self.ltpPubSub = ltpPubSub
        self.respQueue = respQueue
        self.run = True
        self.obj = obj
        self.logger=logger
        self.__minValues = self.__minInterval(self.startTime, self.endTime)
        self.redisConf = redisConf
        self.__toRedis = _toRedis
        self._priorityTicks = _priorityTicks
        self.priorityBroker = ""
    
    def _setPriority(self):
        pass
    
    def _zmq_config(self, host, port):
        context = zmq.Context()
        socket = context.socket(zmq.PUB)
        socket.bind("tcp://{host}:{port}".format(host=host, port=str(port)))
        # socket.setsockopt(zmq.TCP_NODELAY, 1)
        return socket

    def __minInterval(self, start_time, end_time):
        start_time = datetime.datetime.combine(datetime.date(1990,1,1), start_time)
        start_time = start_time + datetime.timedelta(minutes=1)
        end_time = datetime.datetime.combine(datetime.date(1990,1,1), end_time)
        current_time = start_time
        minute_intervals = []

        while current_time <= end_time:
            minute_intervals.append(current_time)
            current_time += datetime.timedelta(minutes=1)
        minute_intervals = [i.time() for i in minute_intervals]
        return minute_intervals
    
    def __connect(self, host, port):
        context = zmq.Context()
        socket = context.socket(zmq.REQ)
        socket.setsockopt( zmq.LINGER,100)
        socket.setsockopt( zmq.AFFINITY,1) 
        socket.setsockopt( zmq.RCVTIMEO, 20000) # 20 Seconds timeout
        socket.connect ("tcp://{host}:{port}".format(host = host, port = port))
        return socket
    
    def __sendReq(self, msg):
        socket = self.__connect(**self.obj.zmqConnection)
        socket.send(json.dumps(msg).encode())
        response = json.loads(socket.recv().decode())
        return response

    def priorityCheck(self, recvMessage = None):
        if self._priorityTicks : 
            
            if self.priorityBroker == "" : 
                self.__delayTrack = {"delay" : {'till' : datetime.datetime.now(), "on" : False}, "wait" : {"till" : "", "on" : False}}
                self.priorityBroker = self.obj._priorityList[0] if len(self.obj._priorityList) != 0 else ""
            
            else: 
                if self.__delayTrack['wait']['on'] : 
                    if datetime.datetime.now() > self.__delayTrack['wait']['till'] : 
                        self.__delayTrack['wait']['on'] = False

                else: 
                    if recvMessage == None :
                        if self.__delayTrack["delay"]["on"] : 
                            if datetime.datetime.now() > self.__delayTrack["delay"]['till'] : 
                                self.__delayTrack["delay"]["on"] = False
                                brokers = self.obj._priorityList
                                idx = brokers.index(self.priorityBroker)
                                self.priorityBroker = brokers[idx + 1] if len(brokers) > idx+1 else brokers[0] 
                                self.logger.critical(f"Shifting Broker to {self.priorityBroker}")
                        else: 
                            self.__delayTrack['delay']['on'] = True
                            self.__delayTrack['delay']['till'] = datetime.datetime.now() + datetime.timedelta(seconds=self.__maxDelay)
                    
                    elif recvMessage != None : 
                        if self.__delayTrack['delay']['on'] : 
                            self.__delayTrack['delay']['on'] = False
    
    def __checkQueue(self):
        while self.run: 
            try: 
                qsize = self.respQueue.qsize()
                for _ in range(qsize):
                    message = self.respQueue.get(timeout=0.1)
                    if message['type'] in ["CLOSE"] :
                        self.logger.critical(f"Connection Closed : {message}")
                        # Restarts the Data websocket as soon as it is closed. 
                        threading.Thread(target = lambda : self.__sendReq({"function": "reconnectWs", 
                                                                        "payload" : {"AccountName" : message['Identifier'].split(":")[1], "spwnID":message['spwID']}
                                                                        }), daemon= True).start()
                        
                    elif message['type'] in ['ERROR']:
                        self.logger.critical(f"Error in Tick Data : {message}") 

                    elif message['type'] == "DATA" : 
                        if self._priorityTicks : 
                            if message['Broker'] == self.priorityBroker : 
                                self.priorityCheck(recvMessage=message)
                                message['data'].update({"broker" : message['Broker']})
                                self.tickQueue.put(message['data'])
                                self.store_MIN(data=message['data'])
                        else:
                            message['data'].update({"broker" : message['Broker']})
                            self.tickQueue.put(message['data'])
                            self.store_MIN(data=message['data'])
                    
                    elif message['type'] == 'GENERAL' : 
                        self.logger.debug(f"General Update : {message}") 
                time.sleep(0.01)
                if qsize == 0 : 
                    self.priorityCheck()
                        
            except queue.Empty:
                self.priorityCheck()
            
            except Exception as e : 
                self.logger.exception(f"Error in Check Queue : {e}")
        
    def _ticksPush(self):
        while self.run:
            try: 
                qsize = self.tickQueue.qsize()
                data = []
                for _ in range(qsize):
                    message = self.tickQueue.get()
                    data.append('{topic} _&_ {messagedata}'.format(topic = f"tick:{message['token']}", 
                                                                            messagedata = json.dumps(message)))
                if data != [] : 
                    self.pubsocket.send_string('{data}'.format(data=data)) #Latency of 12 ms approx
                    
                time.sleep(0.01)
            except: 
                self.logger.exception("Error while pushing ticks")

    def store_MIN(self, data):
        try : 
            minval = int(data['timestamp_str'][-8:-3].replace(":",""))
            if self.MINdata.get(minval) == None :
                self.MINdata[minval] = {}
                if self.prev_minval != minval and minval > self.prev_minval and minval not in self.minvalList:
                    if self.prev_minval != 0 : 
                        self._minPush(self.prev_minval) ### TRY TO PUSH IT ON 59th second.
                    self.premindata = self.MINdata.get(self.prev_minval)
                    self.premindata = {} if self.premindata == None else self.premindata
                    self.prev_minval = minval
                    self.minvalList.append(minval)
                
            token = data['token']
            prev_volume = 0 if self.premindata.get(token) == None else self.premindata.get(token)['v']
            if self.MINdata[minval].get(token) == None: 
                self.MINdata[minval][token] = {"tk" : token, 
                                            "ts" : data['timestamp_str'][:-2] + "00", 
                                            "o" : data['ltp'], 
                                            "h" : data['ltp'], 
                                            "l" : data['ltp'], 
                                            "c" : data['ltp'], 
                                            "v" : data['ttq'], #- prev_volume,
                                            "oi" : 0 if data.get('oi') == None else data.get('oi')}
                self.MINdata[minval][token]['tsp'] = str(int(datetime.datetime.strptime(self.MINdata[minval][token]['ts'], "%Y-%m-%d %H:%M:%S").timestamp()))
            
            ltp = data['ltp'] 
            self.MINdata[minval][token]['c'] = ltp
            self.MINdata[minval][token]['v'] = data['ttq'] #- prev_volume
            
            if ltp > self.MINdata[minval][token]['h']: 
                self.MINdata[minval][token]['h'] = ltp
                
            elif ltp < self.MINdata[minval][token]['l']: 
                self.MINdata[minval][token]['l'] = ltp
            
            if data['oi'] > self.MINdata[minval][token]['oi']:
                self.MINdata[minval][token]['oi'] = data['oi']
                
        except : 
            self.logger.exception(f"Error in store_MIN | data : {data} ")
            
    def runNext(self, currentTime):
        return next((t for t in self.__minValues if t > currentTime), None)
    
    def __connectRedis(self):
        return redis.Redis(connection_pool= self.redispool)
    
    def chunk_list(self, lst, chunk_size):
        """Split a list into chunks of the specified size."""
        return [lst[i:i + chunk_size] for i in range(0, len(lst), chunk_size)]

    def _minPush(self, minval):
        try :
            data = self.MINdata[minval]
            zmqData = ['min:{topic} _&_ {messagedata}'.format(topic=d, messagedata = json.dumps(data[d])) for d in data]
            chunks = self.chunk_list(zmqData, 10)
            for chunk in chunks : 
                self.pubsocket.send_string('{data}'.format(data=chunk))
                
            if self.__toRedis: 
                rcon = self.__connectRedis()
                pipe = rcon.pipeline()
                [pipe.zadd(f"min:{d}", {json.dumps(data[d]) : int(data[d]['tsp'])}) for d in data]
                pipe.execute() 
            self.logger.debug(f"Data pushed into redis at : {datetime.datetime.now()}")
            self.MINdata.pop(minval)
        except :
            self.logger.exception("Error with pushing Min data")

    def __runMin(self):
        nxtrun = self.runNext(datetime.datetime.now().time())
        while self.run:
            if not nxtrun:
                pass
                # self.run = False
            else: 
                curtime = datetime.datetime.now()
                if curtime.time() >= nxtrun:
                    minval = (curtime - datetime.timedelta(minutes = 1)).strftime("%H:%M")
                    data = self.MINdata.get(minval)
                    if data == {} or data == None : 
                        self.logger.critical("ERROR : Data not available to PUSH")
                    else:
                        try: 
                            self.logger.debug(f"Pushing Data at {str(curtime.time())}")
                            self._minPush(minval)
                        except:
                            self.logger.exception("Error with pushing data")

                    nxtrun = self.runNext(datetime.datetime.now().time())
                time.sleep(0.01)

    def StartProcess(self):
        self.pubsocket = self._zmq_config(**self.ltpPubSub)
        self.redispool = redis.ConnectionPool(**self.redisConf)
        threading.Thread(target = self.__checkQueue).start() # Latency to 40ms when data transferred directly from this process.
        threading.Thread(target = self._ticksPush).start() # Latnecy to 15ms when this method is used.
        # threading.Thread(target = self.__runMin).start()

class ZmqDataWs():
    __runServer = True 
    connections = {}
    wsConnections = {}
    respQueue = Queue(maxsize=5000)
    _priorityList = []
        
    def __init__(self, accesstoken, 
                       zmqConnection = {"host" : "127.0.0.1", "port" : "8542"},
                       ltpPubSub = {"host" : "127.0.0.1", "port" : "8543"},
                       ordPubSub = {"host" : "127.0.0.1", "port" : "8544"},
                       __orderUpdates = True,
                       redisConn = {"host" : "127.0.0.1", "port" : "6379"},
                       _toRedis = True,
                       _priorityTicks = False
                       ) :
        
        logger = logging.getLogger(__name__)
        formatter = logging.Formatter('%(asctime)s:%(levelname)s:%(name)s:%(message)s')
        try:
            os.mkdir("./LOGS")
        except:pass
        filehandler = logging.FileHandler('./LOGS/zmqdataws.log')
        filehandler.setFormatter(formatter)
        logger.addHandler(filehandler)
        logger.setLevel(logging.DEBUG)
        self.logger = logger

        self.sparkSession = accesstoken 
        self.spk = SparkLib(access_token=accesstoken)
        self.__functions = {
            "startWS" : self.connect,
            "reconnectWs" : self.reconnect,
            "stopWs" : self.stop,
            "subscribe" : self.subscribe,
            "unsubscribe" : self.unsubscribe,
            "connections" : self.activeConnections
        }
        self.zmqConnection = zmqConnection
        self.ltpPubSub = ltpPubSub
        self._toRedis = _toRedis
        self.redisConn = redisConn
        self._priorityTicks = _priorityTicks
        self.tickHandler = TickHandler(ltpPubSub=self.ltpPubSub, respQueue=self.respQueue, obj=self,
                                       logger=self.logger, redisConf=self.redisConn,_toRedis=self._toRedis, 
                                       _priorityTicks = _priorityTicks)
        self.tickHandler.StartProcess()

        self.__orderUpdates = __orderUpdates
        self.OrderUpdates = OrderWS(sessionid=accesstoken,
                                    ordPubSub=ordPubSub,
                                    logger=self.logger)
        if self.__orderUpdates:
            threading.Thread(target=self.OrderUpdates.connect).start()

    def _middlewareLog(func): 
        @wraps(func)
        def wrapper(self, *args, **kwargs): 
            function_name = func.__name__
            data = args
            if data != () : 
                data = args[0]
                try : 
                    st = time.time()
                    result = func(self, *args, **kwargs)
                    ed = time.time()
                    self.logger.debug(f"ZMQ LOGS | from {str(function_name)} : Payload Received: {str(data)}, response : {str(result)}, timeTaken : {str(round(ed-st,2))}")
                    return result
                except Exception as e : 
                    p = traceback.format_exc()
                    self.logger.exception(f"ZMQ LOGS | from {{str(function_name)}} : Traceback : {p}")
                    return {"status" : False, "error" : True, "message" : f"Error : {str(e)}, Traeback : {p}", "data" : []}
            else: 
                return {"status" : False, "error" : True, "message" : f"Payload not available.", "data" : []}
        return wrapper 

    def __initiateConn(self, AccountName, connId, connType):
        acDict = self.connections[AccountName]
        self.wsConnections[connId] = {
                                    "AccountName" : AccountName, 
                                    "_limit" : acDict['maxTokens'], 
                                    "_usedLimit" : 0, 
                                    "_queue" : Queue() , 
                                    "tokens" : [], 
                                    "_parseTokens" : [], 
                                    "_unsubTokens" : [],
                                    "_broker" : acDict['Broker'],
                                    "_connType" : connType,
                                    "_seg" : acDict['segments']
                                    }     
        if acDict['Broker'] not in self._priorityList:
            self._priorityList.append(acDict['Broker'])  
        self.wsConnections[connId]["_function"] = brokerWS(accountData=acDict['accountData'],
                                                           sparkSession=self.sparkSession,
                                                           respQueue=self.respQueue,
                                                           subsQueue=self.wsConnections[connId]['_queue'],
                                                           connectionType=connType,
                                                           spwID = connId
                                                           )
        self.wsConnections[connId]['_process'] = Process(target= self.wsConnections[connId]['_function'].connect)
        self.wsConnections[connId]['_process'].start()
        # self.wsConnections[connId]['_process']

    def __singleAccount(self, accountName):
        data = self.spk.getOneAccount(accountName)
        if data['status'] and not data['error']:
            return data['data'][0]
        else: 
            return []
        
    @_middlewareLog
    def connect(self, payload):
        if not payload.get("AccountName") : 
            return {"status" : False, "error" : True, "data" : [], "message" : "AccountName missing in payload."} 
        acData = self.__singleAccount(payload['AccountName'])
        if acData == [] : 
            return {"status" : False, "error" : True, "data" : [], "message" : "AccountName not found."} 
        broker = acData['Broker']
        if broker not in BROKERS.keys():
            return {"status" : False, "error" : True, "data" : [], "message" : f"Broker : {broker} not available for ZmqDataWs."}
        elif acData['LastLogin'] != str(datetime.datetime.now().date()) : 
            return {"status" : False, "error" : True, "data" : [], "message" : f"Please generate token for the account : {payload['AccountName']}"}
        
        if payload['AccountName'] not in self.connections.keys() : 
            self.connections[payload['AccountName']] = {
                                                        "accountData" : acData,
                                                        "maxConnections" : BROKERconf[broker]['maxConnection'],
                                                        "curConnections" : 0,
                                                        "maxTokens" : BROKERconf[broker]['tokenLimit'],
                                                        "segments" : BROKERconf[broker]['segments'],
                                                        "connections" : [],
                                                        "Broker" : broker
                                                    }
        connectionType = "ltp" if payload['connectionType'] != "depth" else "depth"
        if connectionType == "depth" : 
            if not BROKERconf[broker]['is_depth'] : 
                return {"status" : False, "error" : True, "message" : f"depth data not supported for Account name : {payload['AccountName']} with broker : {broker}"}

        acDict = self.connections[payload['AccountName']]
        if acDict['curConnections'] + 1 > acDict['maxConnections'] : 
            return {"status" : False, "error" : True, "data":[], 
                    "message" : f"Cannot add new connection on same account name. Current Connection : {acDict['curConnections']}, Max Connections : {acDict['maxConnections']}"}
        
        connIdno = len(acDict['connections']) + 1
        connId = payload['AccountName'] + f"--{str(connIdno)}"
        self.__initiateConn(AccountName=payload['AccountName'], connId= connId, connType=connectionType)
        acDict['connections'].append(connId)
        acDict['curConnections'] += 1
        return {"status" : True, "error" : False, "message" : f"{payload['AccountName']} Connected."}

    def __reconnect(self, accountName, spwId):
        try: 
            con = self.wsConnections[spwId]
            con["_process"].terminate()
            time.sleep(0.5)
            con["_process"].kill()
            con["_process"].close()
            time.sleep(0.2)
            acDict = self.connections[accountName]
            self.wsConnections[spwId]['_queue'] = Queue()
            self.wsConnections[spwId]["_function"] = brokerWS(accountData=acDict['accountData'],
                                                              sparkSession=self.sparkSession,
                                                              respQueue=self.respQueue,
                                                              subsQueue=self.wsConnections[spwId]['_queue'],
                                                              connectionType=self.wsConnections[spwId]['_connType'],
                                                              spwID = spwId
                                                           )
            self.wsConnections[spwId]['_process'] = Process(target= self.wsConnections[spwId]['_function'].connect)
            time.sleep(0.2)
            self.wsConnections[spwId]['_process'].start()
            time.sleep(0.5)
            self.wsConnections[spwId]["_queue"].put({"action" : "subscribe" , 
                                                        "tokens" : self.wsConnections[spwId]['tokens']})
            return True

        except Exception as e :
            self.logger.exception(f"Error with __reconnect : {e}, spw ID : {spwId}, Account Name : {accountName} ")
            return False

    @_middlewareLog
    def reconnect(self, payload):
        if not payload.get("AccountName") : 
            return {"status" : False, "error" : True, "data" : [], "message" : "AccountName missing in payload."} 
        accountName = payload['AccountName']
        if accountName not in self.connections.keys():
            return {"status" : False, "error" : True, "data" : [], "message" : f"AccountName : {accountName} not connected, Unable to reconnect."}
        accountRecon = []
        if payload.get("spwnID") != None : 
            accountRecon.append(self.__reconnect(accountName,payload.get("spwnID")))
        else: 
            spwnIds = self.connections[accountName]['connections']
            for spwId in spwnIds : 
                accountRecon.append(self.__reconnect(accountName,spwId))
        if all(accountRecon) : 
            return {"status": True, "error" : False, "data" : [],  "message" : "Successfully reconnected."}
        else:
            meta_data = {}
            for i in range(len(spwnIds)):
                status = accountRecon[i]
                spwId = spwnIds[i]
                meta_data[spwId] = {"status" : status}
            return {'status' : True, "error" : False, "data" : [meta_data], "message" : "Issue with reconnect."}

    @_middlewareLog
    def stop(self, payload):
        if not payload.get("AccountName") : 
            return {"status" : False, "error" : True, "data" : [], "message" : "AccountName missing in payload."} 
        accountName = payload['AccountName']
        if accountName not in self.connections.keys():
            return {"status" : False, "error" : True, "data" : [], "message" : f"AccountName : {accountName} not connected, Unable to stop."}
        spwnIds = self.connections[accountName]['connections']
        for spwId in spwnIds : 
            con = self.wsConnections[spwId]
            con["_process"].terminate()
            time.sleep(0.2)
            con["_process"].kill()
            con["_process"].close()
            time.sleep(0.2)
            del self.wsConnections[spwId]
        del self.connections[accountName]
        return {"status" : True, "error" : False, "data" : [], "message" : f"Connections closed with account name {accountName}"}

    @_middlewareLog
    def subscribe(self, payload):
        try: 
            if payload.get("AccountName") == None or not payload.get("tokens"): 
                return {"status" : False, "error" : True, "data" : [], "message" : "Invalid Payload."}
            if payload['AccountName'] == "" : 
                accountNames = self.connections.keys()
                brokers = {}
                conntypes = ['ltp', 'depth']
                for ac in accountNames: 
                    bro = self.connections[ac]['Broker']
                    if bro not in brokers: 
                        brokers[bro] = {}
                    brokers[bro][ac] = {"spwid" : self.connections[ac]['connections'], "conntypeData" :  {k : self.wsConnections[k]['_connType'] for k in self.connections[ac]['connections']}} 
                    
                    brokers[bro][ac]["conntypeData"] = {k : [] for k in conntypes}
                    [brokers[bro][ac]["conntypeData"][self.wsConnections[spwid]['_connType']].append(spwid) for spwid in brokers[bro][ac]['spwid']]

                subs = []
                for bro in brokers : 
                    for ctyp in conntypes : 
                        mapToks = []
                        for tk in payload['tokens']: 
                            for ac in brokers[bro]: 
                                for spwId in brokers[bro][ac]['conntypeData'][ctyp] :
                                    conn = self.wsConnections[spwId]
                                    if conn['_usedLimit'] < conn['_limit'] and tk not in mapToks and tk.split(":")[0] in conn['_seg'] :
                                        conn['_usedLimit'] += 1 
                                        conn["_parseTokens"].append(tk)
                                        conn["tokens"].append(tk)
                                        mapToks.append(tk)
                for bro in brokers : 
                    for ac in brokers[bro] : 
                        for spwId in brokers[bro][ac]["spwid"]:
                            conn = self.wsConnections[spwId]
                            if conn["_parseTokens"] != [] : 
                                conn["_queue"].put({"action" : "subscribe" , "tokens" : copy.deepcopy(conn["_parseTokens"])})
                                subs.append({"connection" : spwId, "accountName" : ac, "tokens" : copy.deepcopy(conn["_parseTokens"])})
                            conn["_parseTokens"] = [] 
                return {"status" : True, "error" : False, "data" : subs, "message" : "Tokens subscribed"}           
                                
            else: 
                accountName = payload['AccountName']
                spwIds = self.connections[accountName]['connections']
                conntypeData = {k : self.wsConnections[k]['_connType'] for k in spwIds}
                conntypes = ['ltp', 'depth']
                
                conntypeData = {k : [] for k in conntypes}
                [conntypeData[self.wsConnections[spwid]['_connType']].append(spwid) for spwid in spwIds]
                subs = []
                for ctyp in conntypes : 
                    mapToks = []
                    for tk in payload['tokens'] : 
                        for spwId in conntypeData[ctyp]:
                            conn = self.wsConnections[spwId]
                            if conn['_usedLimit'] < conn['_limit'] and tk not in mapToks and tk.split(":")[0] in conn['_seg'] : #and tk not in conn['tokens']: 
                                conn['_usedLimit'] += 1 
                                conn["_parseTokens"].append(tk)
                                conn["tokens"].append(tk)
                                mapToks.append(tk)

                for spwId in spwIds : 
                    conn = self.wsConnections[spwId]
                    if conn["_parseTokens"] != [] : 
                        conn["_queue"].put({"action" : "subscribe" , "tokens" : copy.deepcopy(conn["_parseTokens"])})
                        subs.append({"connection" : spwId, "accountName" : accountName, "tokens" : copy.deepcopy(conn["_parseTokens"])})
                    conn["_parseTokens"] = [] 
                return {"status" : True, "error" : False, "data" : subs, "message" : "Tokens subscribed"}
        except Exception as e : 
            self.logger.exception("Error with token subscription : ")
            return {"status" : False , "error" : True, "data" : [], "message" : f"Error with token subscription : {str(e)}, Traceback : {traceback.format_exc()}"}
    
    @_middlewareLog
    def unsubscribe(self, payload):
        try: 
            if not payload.get("AccountName") or not payload.get("tokens"): 
                return {"status" : False, "error" : True, "data" : [], "message" : "Invalid Payload."}
            if payload['AccountName'] == "" : 
                accountNames = self.connections.keys()
                unsub = []
                spwIds = []
                [spwIds.extend(self.connections[ac]['connections']) for ac in accountNames]
                print(spwIds)
                for tk in payload['tokens']:
                    for spwId in spwIds :
                        conn = self.wsConnections[spwId]
                        if tk in conn['tokens']: 
                            conn['_unsubTokens'].append(tk)
                            conn['_usedLimit'] -= 1
                            conn['tokens'].remove(tk)
                for spwId in spwIds : 
                    conn = self.wsConnections[spwId]
                    if conn["_unsubTokens"] != [] : 
                        conn["_queue"].put({"action" : "unsubscribe" , "tokens" : copy.deepcopy(conn["_unsubTokens"])})
                        unsub.append({"connection" : spwId, "tokens" : copy.deepcopy(self.conn["_unsubTokens"])})
                    conn["_unsubTokens"] = []
                return {"status" : True, "error" : False, "data" : unsub, "message" : "Tokens unsubscribed"}
            
            else: 
                accountName = payload['AccountName']
                spwIds = self.connections[accountName]['connections']
                unsub = []
                for tk in payload['tokens'] : 
                    for spwId in spwIds :
                        conn = self.wsConnections[spwId]
                        if tk in conn['tokens']: 
                            conn['_unsubTokens'].append(tk)
                            conn['_usedLimit'] -= 1
                            conn['tokens'].remove(tk)
                for spwId in spwIds : 
                    conn = self.wsConnections[spwId]
                    if conn["_unsubTokens"] != [] : 
                        conn["_queue"].put({"action" : "unsubscribe" , "tokens" : copy.deepcopy(conn["_unsubTokens"])})
                        unsub.append({"connection" : spwId, "tokens" : copy.deepcopy(self.conn["_unsubTokens"])})
                    conn["_unsubTokens"] = []
                return {"status" : True, "error" : False, "data" : unsub, "message" : "Tokens unsubscribed"}
        
        except Exception as e : 
            self.logger.exception("Error with token unsubscription : ")
            return {"status" : False , "error" : True, "data" : [], "message" : f"Error with token unsubscription : {str(e)}, Traceback : {traceback.format_exc()}"}

    @_middlewareLog
    def activeConnections(self, payload):
        connDict = {}
        data = self.wsConnections
        for d in data : 
            conndata = data[d]
            connDict[d] = {
                           "AccountName" : conndata['AccountName'], 
                           "_usedLimit" : conndata['_usedLimit'], 
                           "_allowedSeg" : conndata['_seg'], 
                        #    "_tokens" : conndata['tokens'], 
                           "_broker" : conndata['_broker'],
                           "status" : conndata['_process'].is_alive(),
                           "spwID" : d
                           }
        return connDict
    
    def terminateConnections(self):
        self.tickHandler.run = False
        self.OrderUpdates.disconnect()
        for ws in self.wsConnections: 
            con = self.wsConnections[ws]
            con["_process"].terminate()
            time.sleep(0.2)
            con["_process"].kill()
            con["_process"].close()

    def _zmqConnect(self):
        
        context = zmq.Context()
        socket = context.socket(zmq.REP)
        socket.setsockopt(zmq.LINGER, 100)
        socket.bind("tcp://{host}:{port}".format(**self.zmqConnection))
        print("tcp://{host}:{port}".format(**self.zmqConnection))
        
        try: 
            while self.__runServer:
                try: 
                    if socket.poll(1000):
                        message = socket.recv().decode()
                        self.logger.debug(f"Message Received : {message}")
                        message = json.loads(message)
                        if message.get("function") == None or message.get("payload") == None : 
                            resp = {"status" : False, "error" : True, "message" : "Invalid Function or Payload" , "data" : []}
                            socket.send_json(resp)
                        try: 
                            resp = self.__functions[message['function']](message['payload'])
                        except Exception as e : 
                            resp = {"status" : False, "error" : True, "message" : f"{str(e)} : {traceback.format_exc()}"}
                        socket.send_json(resp)

                except KeyboardInterrupt:
                    self.__runServer = False
                
                except zmq.ZMQError as e:
                    self.logger.exception(f"ZMQError : {e}. Exception : ")
                    self.__runServer = False
                    
                except Exception as e : 
                    self.logger.exception("Error received in ZMQ Connect.")
                    socket.send_json({"status" : False, "error" : True, "data" : [], "message" : f"Error : {str(e)}, Traceback : {traceback.format_exc()}"})
                time.sleep(0.01)
        finally : 
            self.terminateConnections()
            socket.close()
            context.term()




import struct
import time
import ssl
import json
import websocket
import os
from threading import Thread

class SmartWebSocketV2(object):
    """
    SmartAPI Web Socket version 2
    """

    ROOT_URI = "wss://smartapisocket.angelone.in/smart-stream"
    HEART_BEAT_MESSAGE = "ping"
    HEART_BEAT_INTERVAL = 10  # Adjusted to 10s
    LITTLE_ENDIAN_BYTE_ORDER = "<"
    # HB_THREAD_FLAG = True

    # Available Actions
    SUBSCRIBE_ACTION = 1
    UNSUBSCRIBE_ACTION = 0

    # Possible Subscription Mode
    LTP_MODE = 1
    QUOTE = 2
    SNAP_QUOTE = 3
    DEPTH = 4

    # Exchange Type
    NSE_CM = 1
    NSE_FO = 2
    BSE_CM = 3
    BSE_FO = 4
    MCX_FO = 5
    NCX_FO = 7
    CDE_FO = 13

    # Subscription Mode Map
    SUBSCRIPTION_MODE_MAP = {
        1: "LTP",
        2: "QUOTE",
        3: "SNAP_QUOTE",
        4: "DEPTH"
    }

    wsapp = None
    input_request_dict = {}
    current_retry_attempt = 0

    def __init__(self, auth_token, api_key, client_code, feed_token, max_retry_attempt=1,
                 on_Message=None, on_Error=None, on_Close=None, on_Open=None
                 ):
        """
            Initialise the SmartWebSocketV2 instance
            Parameters
            ------
            auth_token: string
                jwt auth token received from Login API
            api_key: string
                api key from Smart API account
            client_code: string
                angel one account id
            feed_token: string
                feed token received from Login API
        """
        self.auth_token = auth_token
        self.api_key = api_key
        self.client_code = client_code
        self.feed_token = feed_token
        self.last_pong_timestamp = None
        self.MAX_RETRY_ATTEMPT = max_retry_attempt
        self.retry_delay = 10
        self.retry_duration = 60

        self.on_data = on_Message
        self.on_error = on_Error
        self.on_close = on_Close
        self.on_open = on_Open

        if not self._sanity_check():
            raise Exception("Provide valid value for all the tokens")

    def _sanity_check(self):
        if not all([self.auth_token, self.api_key, self.client_code, self.feed_token]):
            return False
        return True

    def _on_message(self, wsapp, message):
        if message != "pong":
            parsed_message = self._parse_binary_data(message)
            # Check if it's a control message (e.g., heartbeat)
            if self._is_control_message(parsed_message):
                self._handle_control_message(parsed_message)
            else:
                self.on_data(parsed_message)
        else:
            self.on_data(message)

    def _is_control_message(self, parsed_message):
        return "subscription_mode" not in parsed_message

    def _handle_control_message(self, parsed_message):
        if parsed_message["subscription_mode"] == 0:
            self._on_pong(self.wsapp, "pong")
        elif parsed_message["subscription_mode"] == 1:
            self._on_ping(self.wsapp, "ping")
        # Invoke on_control_message callback with the control message data
        if hasattr(self, 'on_control_message'):
            self.on_control_message(self.wsapp, parsed_message)

    def _on_data(self, wsapp, data, data_type, continue_flag):
        if data_type == 2:
            parsed_message = self._parse_binary_data(data)
            self.on_data(parsed_message)

    def _on_open(self, wsapp):
        try:
            if self.on_open:
                self.on_open()
        except Exception as e:
            print("Error in on_open function: %s", e)
        self.current_retry_attempt = 0


    def subscribe(self, correlation_id, mode, token_list):
        """
            This Function subscribe the price data for the given token
            Parameters
            ------
            correlation_id: string
                A 10 character alphanumeric ID client may provide which will be returned by the server in error response
                to indicate which request generated error response.
                Clients can use this optional ID for tracking purposes between request and corresponding error response.
            mode: integer
                It denotes the subscription type
                possible values -> 1, 2 and 3
                1 -> LTP
                2 -> Quote
                3 -> Snap Quote
            token_list: list of dict
                Sample Value ->
                    [
                        { "exchangeType": 1, "tokens": ["10626", "5290"]},
                        {"exchangeType": 5, "tokens": [ "234230", "234235", "234219"]}
                    ]
                    exchangeType: integer
                    possible values ->
                        1 -> nse_cm
                        2 -> nse_fo
                        3 -> bse_cm
                        4 -> bse_fo
                        5 -> mcx_fo
                        7 -> ncx_fo
                        13 -> cde_fo
                    tokens: list of string
        """
        try:
            request_data = {
                "correlationID": correlation_id,
                "action": self.SUBSCRIBE_ACTION,
                "params": {
                    "mode": mode,
                    "tokenList": token_list
                }
            }
            if mode == 4:
                for token in token_list:
                    if token.get('exchangeType') != 1:
                        error_message = f"Invalid ExchangeType:{token.get('exchangeType')} Please check the exchange type and try again it support only 1 exchange type"
                        raise ValueError(error_message)

            if self.input_request_dict.get(mode) is None:
                self.input_request_dict[mode] = {}

            for token in token_list:
                if token['exchangeType'] in self.input_request_dict[mode]:
                    self.input_request_dict[mode][token['exchangeType']].extend(token["tokens"])
                else:
                    self.input_request_dict[mode][token['exchangeType']] = token["tokens"]

            if mode == self.DEPTH:
                total_tokens = sum(len(token["tokens"]) for token in token_list)
                quota_limit = 50
                if total_tokens > quota_limit:
                    error_message = f"Quota exceeded: You can subscribe to a maximum of {quota_limit} tokens only."
                    raise Exception(error_message)

            self.wsapp.send(json.dumps(request_data))

        except Exception as e:
            raise e

    def unsubscribe(self, correlation_id, mode, token_list):
        """
            This function unsubscribe the data for given token
            Parameters
            ------
            correlation_id: string
                A 10 character alphanumeric ID client may provide which will be returned by the server in error response
                to indicate which request generated error response.
                Clients can use this optional ID for tracking purposes between request and corresponding error response.
            mode: integer
                It denotes the subscription type
                possible values -> 1, 2 and 3
                1 -> LTP
                2 -> Quote
                3 -> Snap Quote
            token_list: list of dict
                Sample Value ->
                    [
                        { "exchangeType": 1, "tokens": ["10626", "5290"]},
                        {"exchangeType": 5, "tokens": [ "234230", "234235", "234219"]}
                    ]
                    exchangeType: integer
                    possible values ->
                        1 -> nse_cm
                        2 -> nse_fo
                        3 -> bse_cm
                        4 -> bse_fo
                        5 -> mcx_fo
                        7 -> ncx_fo
                        13 -> cde_fo
                    tokens: list of string
        """
        try:
            request_data = {
                "correlationID": correlation_id,
                "action": self.UNSUBSCRIBE_ACTION,
                "params": {
                    "mode": mode,
                    "tokenList": token_list
                }
            }
            self.input_request_dict.update(request_data)
            self.wsapp.send(json.dumps(request_data))
        except Exception as e:
            raise e

    def wsapp_run_forever(self):
        self.wsapp.run_forever(sslopt={"cert_reqs": ssl.CERT_NONE}, ping_interval=self.HEART_BEAT_INTERVAL,
                               ping_payload=self.HEART_BEAT_MESSAGE)#,ping_timeout=self.HEART_BEAT_INTERVAL-1)

    def connect(self):
        """
            Make the web socket connection with the server
        """
        headers = {
            "Authorization": self.auth_token,
            "x-api-key": self.api_key,
            "x-client-code": self.client_code,
            "x-feed-token": self.feed_token
        }

        try:
            self.wsapp = websocket.WebSocketApp(self.ROOT_URI, header=headers, on_open=self._on_open,
                                                on_error=self._on_error, on_close=self._on_close, on_data=self._on_data,
                                                )

            self.ws_thread = Thread(target=self.wsapp_run_forever)
            self.ws_thread.daemon = False
            self.ws_thread.start()
        except Exception as e:
            print(f"Error occurred during WebSocket connection: {e}")
            raise e

    def close_connection(self):
        """
        Closes the connection
        """
        if self.wsapp:
            self.wsapp.close()

    def _on_error(self, wsapp, error):
        try:
            if self.on_error:
                self.on_error(error)
        except Exception as e:
            print("Error in on_error function: %s", e)
        self.retry_connect()

    def retry_connect(self):
        if self.current_retry_attempt < self.MAX_RETRY_ATTEMPT:
            print(f"Retrying connection for {self.current_retry_attempt+1} of {self.MAX_RETRY_ATTEMPT}")
            time.sleep(self.retry_delay)
            self.current_retry_attempt += 1
            self.connect()
        else:
            print("Max retry attempts reached.")

    def _on_close(self, wsapp, close_status_code, close_msg):
        self.on_close(close_status_code, close_msg)

    def _parse_binary_data(self, binary_data):
        parsed_data = {
            "subscription_mode": self._unpack_data(binary_data, 0, 1, byte_format="B")[0],
            "exchange_type": self._unpack_data(binary_data, 1, 2, byte_format="B")[0],
            "token": SmartWebSocketV2._parse_token_value(binary_data[2:27]),
            "sequence_number": self._unpack_data(binary_data, 27, 35, byte_format="q")[0],
            "exchange_timestamp": self._unpack_data(binary_data, 35, 43, byte_format="q")[0],
            "last_traded_price": self._unpack_data(binary_data, 43, 51, byte_format="q")[0]
        }
        try:
            parsed_data["subscription_mode_val"] = self.SUBSCRIPTION_MODE_MAP.get(parsed_data["subscription_mode"])

            if parsed_data["subscription_mode"] in [self.QUOTE, self.SNAP_QUOTE]:
                parsed_data["last_traded_quantity"] = self._unpack_data(binary_data, 51, 59, byte_format="q")[0]
                parsed_data["average_traded_price"] = self._unpack_data(binary_data, 59, 67, byte_format="q")[0]
                parsed_data["volume_trade_for_the_day"] = self._unpack_data(binary_data, 67, 75, byte_format="q")[0]
                parsed_data["total_buy_quantity"] = self._unpack_data(binary_data, 75, 83, byte_format="d")[0]
                parsed_data["total_sell_quantity"] = self._unpack_data(binary_data, 83, 91, byte_format="d")[0]
                parsed_data["open_price_of_the_day"] = self._unpack_data(binary_data, 91, 99, byte_format="q")[0]
                parsed_data["high_price_of_the_day"] = self._unpack_data(binary_data, 99, 107, byte_format="q")[0]
                parsed_data["low_price_of_the_day"] = self._unpack_data(binary_data, 107, 115, byte_format="q")[0]
                parsed_data["closed_price"] = self._unpack_data(binary_data, 115, 123, byte_format="q")[0]

            if parsed_data["subscription_mode"] == self.SNAP_QUOTE:
                parsed_data["last_traded_timestamp"] = self._unpack_data(binary_data, 123, 131, byte_format="q")[0]
                parsed_data["open_interest"] = self._unpack_data(binary_data, 131, 139, byte_format="q")[0]
                parsed_data["open_interest_change_percentage"] = self._unpack_data(binary_data, 139, 147, byte_format="q")[0]
                parsed_data["upper_circuit_limit"] = self._unpack_data(binary_data, 347, 355, byte_format="q")[0]
                parsed_data["lower_circuit_limit"] = self._unpack_data(binary_data, 355, 363, byte_format="q")[0]
                parsed_data["52_week_high_price"] = self._unpack_data(binary_data, 363, 371, byte_format="q")[0]
                parsed_data["52_week_low_price"] = self._unpack_data(binary_data, 371, 379, byte_format="q")[0]
                best_5_buy_and_sell_data = self._parse_best_5_buy_and_sell_data(binary_data[147:347])
                parsed_data["best_5_buy_data"] = best_5_buy_and_sell_data["best_5_sell_data"]
                parsed_data["best_5_sell_data"] = best_5_buy_and_sell_data["best_5_buy_data"]

            if parsed_data["subscription_mode"] == self.DEPTH:
                parsed_data.pop("sequence_number", None)
                parsed_data.pop("last_traded_price", None)
                parsed_data.pop("subscription_mode_val", None)
                parsed_data["packet_received_time"]=self._unpack_data(binary_data, 35, 43, byte_format="q")[0]
                depth_data_start_index = 43
                depth_20_data = self._parse_depth_20_buy_and_sell_data(binary_data[depth_data_start_index:])
                parsed_data["depth_20_buy_data"] = depth_20_data["depth_20_buy_data"]
                parsed_data["depth_20_sell_data"] = depth_20_data["depth_20_sell_data"]

            return parsed_data
        except Exception as e:
            print(f"Error occurred during binary data parsing: {e}")
            raise e

    def _unpack_data(self, binary_data, start, end, byte_format="I"):
        """
            Unpack Binary Data to the integer according to the specified byte_format.
            This function returns the tuple
        """
        return struct.unpack(self.LITTLE_ENDIAN_BYTE_ORDER + byte_format, binary_data[start:end])

    @staticmethod
    def _parse_token_value(binary_packet):
        token = ""
        for i in range(len(binary_packet)):
            if chr(binary_packet[i]) == '\x00':
                return token
            token += chr(binary_packet[i])
        return token

    def _parse_best_5_buy_and_sell_data(self, binary_data):

        def split_packets(binary_packets):
            packets = []

            i = 0
            while i < len(binary_packets):
                packets.append(binary_packets[i: i + 20])
                i += 20
            return packets

        best_5_buy_sell_packets = split_packets(binary_data)

        best_5_buy_data = []
        best_5_sell_data = []

        for packet in best_5_buy_sell_packets:
            each_data = {
                "flag": self._unpack_data(packet, 0, 2, byte_format="H")[0],
                "quantity": self._unpack_data(packet, 2, 10, byte_format="q")[0],
                "price": self._unpack_data(packet, 10, 18, byte_format="q")[0],
                "no of orders": self._unpack_data(packet, 18, 20, byte_format="H")[0]
            }

            if each_data["flag"] == 0:
                best_5_buy_data.append(each_data)
            else:
                best_5_sell_data.append(each_data)

        return {
            "best_5_buy_data": best_5_buy_data,
            "best_5_sell_data": best_5_sell_data
        }

    def _parse_depth_20_buy_and_sell_data(self, binary_data):
        depth_20_buy_data = []
        depth_20_sell_data = []

        for i in range(20):
            buy_start_idx = i * 10
            sell_start_idx = 200 + i * 10

            # Parse buy data
            buy_packet_data = {
                "quantity": self._unpack_data(binary_data, buy_start_idx, buy_start_idx + 4, byte_format="i")[0],
                "price": self._unpack_data(binary_data, buy_start_idx + 4, buy_start_idx + 8, byte_format="i")[0],
                "num_of_orders": self._unpack_data(binary_data, buy_start_idx + 8, buy_start_idx + 10, byte_format="h")[0],
            }

            # Parse sell data
            sell_packet_data = {
                "quantity": self._unpack_data(binary_data, sell_start_idx, sell_start_idx + 4, byte_format="i")[0],
                "price": self._unpack_data(binary_data, sell_start_idx + 4, sell_start_idx + 8, byte_format="i")[0],
                "num_of_orders": self._unpack_data(binary_data, sell_start_idx + 8, sell_start_idx + 10, byte_format="h")[0],
            }

            depth_20_buy_data.append(buy_packet_data)
            depth_20_sell_data.append(sell_packet_data)

        return {
            "depth_20_buy_data": depth_20_buy_data,
            "depth_20_sell_data": depth_20_sell_data
        }

    # def on_message(self, wsapp, message):
    #     print('Received message: ', message)
    #
    # def on_data(self, wsapp, data):
    #     print('Received data: ', data)
    #
    # def on_control_message(self, wsapp, message):
    #     print('Received control message: ', message)
    #
    # def on_close(self, wsapp, close_status_code, close_msg):
    #     print('Connection closed')
    #
    # def on_open(self, wsapp):
    #     print('Connection opened')
    #
    # def on_error(self, error):
    #     print('Error occurred')

if __name__ == "__main__":
    AUTH_TOKEN = 'eyJhbGciOiJIUzUxMiJ9.eyJ1c2VybmFtZSI6Ik4yNzg3MTgiLCJyb2xlcyI6MCwidXNlcnR5cGUiOiJVU0VSIiwidG9rZW4iOiJleUpoYkdjaU9pSlNVekkxTmlJc0luUjVjQ0k2SWtwWFZDSjkuZXlKMWMyVnlYM1I1Y0dVaU9pSmpiR2xsYm5RaUxDSjBiMnRsYmw5MGVYQmxJam9pZEhKaFpHVmZZV05qWlhOelgzUnZhMlZ1SWl3aVoyMWZhV1FpT2pRc0luTnZkWEpqWlNJNklqTWlMQ0prWlhacFkyVmZhV1FpT2lJd1pXSXlPR05tWWkweFpqSXdMVE13WW1ZdFltSXpNeTB6T1RrMll6azVPRFV6TldVaUxDSnJhV1FpT2lKMGNtRmtaVjlyWlhsZmRqRWlMQ0p2Ylc1bGJXRnVZV2RsY21sa0lqbzBMQ0p3Y205a2RXTjBjeUk2ZXlKa1pXMWhkQ0k2ZXlKemRHRjBkWE1pT2lKaFkzUnBkbVVpZlgwc0ltbHpjeUk2SW5SeVlXUmxYMnh2WjJsdVgzTmxjblpwWTJVaUxDSnpkV0lpT2lKT01qYzROekU0SWl3aVpYaHdJam94TnpFNU5EY3pOakUxTENKdVltWWlPakUzTVRrek56WTVNamNzSW1saGRDSTZNVGN4T1RNM05qa3lOeXdpYW5ScElqb2lORGhoWkRjM01EVXRabU5rTWkwMFlqQTJMVGxsTm1VdE1UUmlaRFV6TmpKallXRmhJbjAuU2RJUFI4OWk3VnQ0UlA0TExFelpSaTdSMG94VnlzVHhXbjlXeW5mQUhUaTZ0OVVzdjNrV2VVNkVubUpxWWpLdXVCaklfbVg3clM5aThWNk5jdWFOMHVESm85SVJCRVlZRWJJZDVfZmtWay1STU1MbWhrN093UWVMdk90QnJvRllhaGVDRG5NbjR1UTJzQ3Z4Z1VWZUJkTTNTeFY2dUxXd2JSdlFGa01nMzRBIiwiQVBJLUtFWSI6IlZhQXJkak1CIiwiaWF0IjoxNzE5Mzc2OTg3LCJleHAiOjE3MTk0NzM2MTV9.JjHFs-ILCP0ZFUsYewNdufrqokLXemawEjhWMU4utqDa9PboiaWsySfYC32qKL6vnSUzyTZavh_wpPtBDe4q4Q'
    API_KEY = 'VaArdjMB'
    CLIENT_CODE = 'N278718'
    FEED_TOKEN    = 'eyJhbGciOiJIUzUxMiJ9.eyJ1c2VybmFtZSI6Ik4yNzg3MTgiLCJpYXQiOjE3MTkzNzY5ODcsImV4cCI6MTcxOTQ2MzM4N30.76Z3LZmCmB4IdVO8uL3Qw8_2H2zlvIc7x-uR4Lqc08rBjXh2i_yTOZZinNlWy4QlWDIpYOp5_OSKAcvSzjXb9A'
    correlation_id = "abc123"
    action = 1
    # ACTION:
        # 1 : subscribe
        # 0 : unsubscribe
        
    mode = 1
    # modes : 
    # 1 (LTP)
    # 2 (Quote)
    # 3 (Snap Quote)
    # 4 (20-Depth)
    
    token_list = [
        {
            "exchangeType": 1,
            "tokens": ["26009","1594"]
        }
    ]
    token_list1 = [
        {
            "action": 0,
            "exchangeType": 1,
            "tokens": ["26009"]
        }
    ]
    
    def on_message(message):
        print("Ticks: {}".format(message))

    def on_open():
        print("on open")
        sws.subscribe(correlation_id, mode, token_list)
        # sws.unsubscribe(correlation_id, mode, token_list1)

    def on_error(error):
        print(f'Error : {error}')

    def on_close(close_status_code, close_msg):
        print(f"Closed : {close_status_code} : {close_msg}")

    sws = SmartWebSocketV2(AUTH_TOKEN, API_KEY, CLIENT_CODE, FEED_TOKEN,max_retry_attempt=3,
                           on_Message=on_message, on_Error=on_error, on_Close=on_close, on_Open=on_open)

    sws.connect()
    time.sleep(40)
    sws.close_connection()




# -*- coding: utf-8 -*-
"""
Created on Thu Dec 14 13:32:00 2023

@author: Niraj
"""

import requests
import base64
import jwt
import json

# nse_cm - NSE
# bse_cm - BSE
# nse_fo - NFO
# bse_fo - BFO
# cde_fo - CDS

# NRML - Normal
# CNC - Cash and Carry
# MIS - MIS
# CO - Cover Order

# L - Limit
# MKT - Market
# SL - Stop loss limit
# SL-M - Stop loss market

class KotakApi():
    ENDPOINTS = {"AccessToken" : "https://napi.kotaksecurities.com/oauth2/token",
                 "login.validate" : "/login/1.0/login/v2/validate", 
                 "login.generate" : "/login/1.0/login/otp/generate",
                 "login.session" : "/login/1.0/login/v2/validate",
                 "script-master" : "/Files/1.0/masterscrip/v1/file-paths",
                 "order.place" : "/Orders/2.0/quick/order/rule/ms/place?sId={sId}",
                 "order.modify" : "/Orders/2.0/quick/order/vr/modify?sId={sId}",
                 "order.cancel" : "/Orders/2.0/quick/order/cancel?sId={sId}serverId",
                 "order.book" : "/Orders/2.0/quick/user/orders?sId={sId}&hsServerId={hsServerId}",
                 "trade.book" : "/Orders/2.0/quick/user/trades?sId={hsServerId}",
                "position" : "/Orders/2.0/quick/user/positions?sId={hsServerId}",
                 "portfolio" : "/Portfolio/1.0/portfolio/v1/holdings?alt={alt}",
                 "limits" : "/Orders/2.0/quick/user/limits?sId={hsServerId}",
                 "margin" : "/Orders/2.0/quick/user/check-margin?sId={hsServerId}",
                 }
    BASEURL = "https://gw-napi.kotaksecurities.com"
    
    def __init__(self, consumerKey = None, 
                       secretKey = None, 
                       usernameAPI = None, 
                       passwordAPI = None, 
                       phoneNo = None, 
                       password = None, 
                       loginDetails = None):

        self.consumerKey = consumerKey
        self.secretKey = secretKey
        self.passwordAPI = passwordAPI
        self.usernameAPI = usernameAPI
        self.phoneNo = phoneNo
        self.password = password

        self.neofinKey = "neotradeapi"
        self.reqsession = requests.Session()
        self.header = {
                "Accept" : "*/*",
                "Accept-Encoding" : "gzip, deflate, br",
                "Connection" : "keep-alive", 
                "Version" : "2.1",
                "Content-Type" : "application/x-www-form-urlencoded"
            }

        if loginDetails != None:
            self.token = loginDetails['token']
            self.userID = loginDetails['userID']
            self.sid = loginDetails['sid']
            self.hsServerId = loginDetails['hsServerId']
            self.header['Sid'] = self.sid
            self.header['Auth'] = self.token
            self.header['neo-fin-key'] = self.neofinKey
            self.header['Authorization'] = loginDetails['Authorization']

    def get_login_details(self):
        return {"token" : self.header['Auth'], "userID" : self.userID, "sid" : self.sid, "hsServerId" : self.hsServerId, "Authorization" : self.header['Authorization']}

    def accessToken(self, secretKey = None, usernameAPI = None, passwordAPI = None):
        secretKey = secretKey if self.secretKey == None else self.secretKey
        username = usernameAPI if self.usernameAPI == None else self.usernameAPI
        password = passwordAPI if self.passwordAPI == None else self.passwordAPI
        data = ":".join([self.consumerKey, secretKey])
        encodedKeys = base64.b64encode(data.encode()).decode()
        headers = {"Authorization" : f"Basic {encodedKeys}"}
        body = {"grant_type": "client_credentials", "username" : username, "password" : password}
        data = self._requestAPI("POST", self.ENDPOINTS['AccessToken'], body = body, header = headers)
        if data['status'] and not data['error']:
            try: 
                if data['data'][0].get("access_token") != None:
                    self.access_token = data['data'][0]['access_token']
                    self.header['Authorization'] = f"Bearer {self.access_token}"
                    # self.authParams[]
                    data['message'] = "Access Token generated."
                    return data
                else: 
                    data['status'] = False
                    data['error'] = True
                    data['message'] = "Access Token not found"
                    return data
            except: 
                data['status'] = False
                data['error'] = True
                data['message'] = "Access Token not found"
                return data
        else : 
            data['message'] = "Access Token not found"
            return data
        
    def __decodeJWT(self, token):
        return jwt.decode(token, options= {"verify_signature": False})
    
    def loginValidate(self, phoneNo = None, password = None):
        phoneNo = phoneNo if self.phoneNo == None else self.phoneNo
        password = password if self.password == None else self.password
        phoneNo = "+91" + str(phoneNo) if str(phoneNo)[:3] != "+91" else str(phoneNo)
        body = {
                "mobileNumber": phoneNo,
                "password": password,
            }
        url = "".join([self.BASEURL, self.ENDPOINTS['login.validate']])
        data = self._requestAPI("POST", url, body = body)
        if data['status'] and not data['error']:
            try: 
                if data['data'][0].get("data"):
                    respData = data['data'][0]['data']
                    self.token = respData['token']
                    self.userID = self.__decodeJWT(self.token)['sub']
                    self.sid = respData['sid']
                    self.header['Sid'] = self.sid
                    self.header['Auth'] = self.token
                    self.header['neo-fin-key'] = "neotradeapi"
                    data['message'] = "Login validation complete"
                    return data
                else: 
                    data['status'] = False
                    data['error'] = True
                    data['message'] = "Login validation failed"
                    return data
            except: 
                data['status'] = False
                data['error'] = True
                data['message'] = "Login validation failed"
                return data
        else: 
            data['message'] = "Login validation failed"
            return data
    
    def loginGenerate(self):
        body = {
                "userId": self.userID,
                "sendEmail": True,
                "isWhitelisted": True
            }
        url = "".join([self.BASEURL, self.ENDPOINTS['login.generate']])
        data = self._requestAPI("POST", url, body = body)
        try : 
            if data['data'][0].get('error') :
                data['message'] = f"Error {data['data'][0]['error']['message']}, Code {data['data'][0]['error']['code']}"
                
            else: 
                data['message'] = "OTP generated"
                return data
        except: 
            data['message'] = "Otp not generated."
            return data

    def loginSession(self, otp):
        if len(str(otp)) == 4 :
            body = {
                "userId": self.userID,
                "otp": str(otp)
            }
        elif len(str(otp)) == 6 :
            body = {
                "userId": self.userID,
                "mpin": str(otp)
            }
        else:
            return {"status" : False, "error" : True, "message" : "Invalid OTP"}
        url = "".join([self.BASEURL, self.ENDPOINTS['login.session']])
        data = self._requestAPI("POST", url, body = body)
        if data['status'] and not data['error'] : 
            try: 
                if data['data'][0].get('data') : 
                    self.hsServerId = data['data'][0]['data']['hsServerId']
                    self.header['Auth'] = data['data'][0]['data']['token']
                    data['message'] = "Login Successful"
                    return data
                else: 
                    data['status'] = False
                    data['error'] = True
                    data['message'] = "Login Validation failed"
                    return data
                
            except: 
                data['status'] = False
                data['error'] = True
                data['message'] = "Login Validation failed"
                return data
            
        else: 
            data['message'] = "Login Validation failed"
            return data

    def scriptMaster(self):
        url = "".join([self.BASEURL, self.ENDPOINTS['script-master']])
        return self._requestAPI("GET", url)
        
    def placeorder(self, pr, es, pc, qt, tt, pt, ts, am = "NO", tp = 0,dq = 0, mp = 0, pf = "N", rt = "DAY",):
        """
        Parameters
        ----------
        pr : float
            The price to execute the order Ex:2000.00, 250.50, etc.
        es : str
            exchange segment - Eg:nse_cm -> NSE, bse_cm -> BSE, nse_fo -> NFO, bse_fo -> BFO, cde_fo -> CDS, bcs_fo -> BCD, mcx_fo -> MCX .
        pc : str
            Product code (NRML -> Normal, CNC -> Cash and Carry, MIS ->Margin Intraday Squareoff for futures and options, INTRADAY -> INTRADAY, CO->Cover Order, BO -> Bracket Order, PRIME -> Custom product) .
        qt : int
            Quantity to transact (Ex:10, 20, etc) .
        tt : str
            Transaction Type (Buy -> B or Sell -> S) .
        pt : str
            Order Type (L -> Limit, MKT -> Market, SL -> Stop loss limit, SL-M -> Stop loss market) .
        ts : str
            Exchange trading symbol of the instrument Ex:TCS-EQ, ACC-EQ, etc.
        am : str, optional
            After Market Order(Allowed Values -> YES/NO) . The default is "NO".
        dq : int, optional
            Disclosed quantity (Quantity to disclose publicly (for equity trades)) Ex:5, 10 etc . The default is 0.
        mp : float, optional
            Market protection. 0 by default (Ex: 5.00, 8.00 etc) . The default is 0.
        pf : str, optional
            PosSqrFlg (Allowed Values -> Y/N, No Need to change this keep this as N). The default is "N".
        rt : str, optional
            Order Duration (Allowed values - either DAY -> Regular order or IOC -> Immediate or Cancel) . The default is "DAY".
        tp : float
            trigger price
        
        Returns
        -------
        Order ID.

        """
        body = locals()
        del (body["self"])
        body = {"jData" : json.dumps(body)}
        url = "".join([self.BASEURL, self.ENDPOINTS['order.place'].format(sId = self.hsServerId)])
        header = self.header
        header["Content-Type"] = "application/x-www-form-urlencoded"
        data = self._requestAPI("POST", url, data = body,header= header)
        if data['status'] and not data['error'] : 
            try: 
                if data['data'][0].get('nOrdNo') == None :
                    data['status'] = False
                    data['error'] = True
                    data['message'] = data['data'][0]['errMsg']
                else: 
                    data['message'] = "Order Placed Successfully"
            except: 
                data['status'] = False
                data['error'] = True
                data['message'] = "Error in placing order"
        else: 
            try: 
                data['message'] = data['data'][0]['errMsg']
            except : 
                data['message'] = "Unable to Place order"
        return data

    def modifyorder(self, tk, pc, ts, tt, pr, qt, no, es, pt, tp = '0', mp = '0', dd = "NA", dq = '0', vd = "DAY",am = "NO"):
        """
        Parameters
        ----------
        tk : str
            Instrument Token
        pc : str
            Product code (NRML -> Normal, CNC -> Cash and Carry, MIS ->Margin Intraday Squareoff for futures and options, INTRADAY -> INTRADAY, CO->Cover Order, BO -> Bracket Order, PRIME -> Custom product).
        ts : str
            Trading Symbol.
        tt : str
            Transaction Type
        pr : float
            The price to execute the order Ex:2000.00, 250.50, etc.
        qt : int
            Quantity 
        tp : float
            Trigger Price
        no : int
            Order Number
        es : str
            exchange segment - Eg:nse_cm -> NSE, bse_cm -> BSE, nse_fo -> NFO, bse_fo -> BFO, cde_fo -> CDS, bcs_fo -> BCD, mcx_fo -> MCX .
        pt : str
            Order Type (L -> Limit, MKT -> Market, SL -> Stop loss limit, SL-M -> Stop loss market) .
        Returns
        -------
        Order ID

        """
        # fq, am
        qt = str(qt)
        body = locals()
        del (body["self"])
        body = {"jData" : json.dumps(body)}
        url = "".join([self.BASEURL, self.ENDPOINTS['order.modify'].format(sId = self.hsServerId)])  
        header = self.header
        header["Content-Type"] = "application/x-www-form-urlencoded"
        data = self._requestAPI("POST", url, data = body,header= header)
        if data['status'] and not data['error'] : 
            try: 
                if data['data'][0].get('nOrdNo') == None :
                    data['status'] = False
                    data['error'] = True
                    data['message'] = data['data'][0]['message']
                else: 
                    data['message'] = "Order Modified Successfully"
            except: 
                data['status'] = False
                data['error'] = True
                data['message'] = "Error in modifying order"
        else: 
            try: 
                data['message'] = data['data'][0]['errMsg']
            except : 
                data['message'] = "Unable to modify order"
                
        return data
    
    def cancelorder(self, on, am = "NO"):
        body = {"on" : on}
        body = {"jData" : json.dumps(body)}
        url = "".join([self.BASEURL, self.ENDPOINTS['order.cancel'].format(sId = self.hsServerId)])  
        header = self.header
        header["Content-Type"] = "application/x-www-form-urlencoded"
        data = self._requestAPI("POST", url, data = body,header= header)
        if data['status'] and not data['error'] : 
            try: 
                if data['data'][0].get('result') == None :
                    data['status'] = False
                    data['error'] = True
                    data['message'] = data['data'][0]['message']
                else: 
                    data['message'] = "Order Cancelled Successfully"
            except: 
                data['status'] = False
                data['error'] = True
                data['message'] = "Error in cancelling order"
        else: 
            try: 
                data['message'] = data['data'][0]['errMsg']
            except : 
                data['message'] = "Unable to Cancel order"
                
        return data
    
    def orderbook(self):
        url = "".join([self.BASEURL, self.ENDPOINTS['order.book'].format(sId = self.sid, hsServerId = self.hsServerId)])
        data =  self._requestAPI("GET", url)
        if data['data'][0].get("errMsg") == None : 
            data['message'] = "Data Received"
        else: 
            data['message'] = data['data'][0].get("errMsg")
            data['status'] = False
            data['error'] = True
        return data

    def tradebook(self):
        url = "".join([self.BASEURL, self.ENDPOINTS['trade.book'].format(hsServerId = self.hsServerId)])
        data = self._requestAPI("GET", url)
        if data['data'][0].get("errMsg") == None : 
            data['message'] = "Data Received"
        else: 
            data['message'] = data['data'][0].get("errMsg")
            data['status'] = False
            data['error'] = True
        return data

    def position(self):
        url = "".join([self.BASEURL, self.ENDPOINTS['position'].format(hsServerId = self.hsServerId)])
        return self._requestAPI("GET", url)

    def portfolio(self, alt = "false"):
        url = "".join([self.BASEURL, self.ENDPOINTS['portfolio'].format(alt = alt)])
        return self._requestAPI("GET", url)

    def limits(self,segment="ALL", exchange="ALL", product="ALL"):
        payload = {'segment': segment, 'exchange': exchange, 'product': product}
        header_params = {'Authorization':self.header['Authorization'],
                         "Sid": self.sid,
                         "Auth": self.token,
                         "neo-fin-key": self.neofinKey,
                         "accept": "application/json",
                         "Content-Type": "application/x-www-form-urlencoded"}
        url = "".join([self.BASEURL, self.ENDPOINTS['limits'].format(hsServerId = self.hsServerId)])
        return self._requestAPI("POST", url, data = payload, header = header_params)

    def margin(self,exchange_segment, price, order_type, product, quantity, instrument_token, transaction_type,
               trigger_price, broker_name, branch_id, stop_loss_type, stop_loss_value,
               square_off_type, square_off_value, trailing_stop_loss, trailing_sl_value):
        header_params = {'Authorization':self.header['Authorization'],
                         "Sid": self.sid,
                         "Auth": self.token,
                         "neo-fin-key": self.neofinKey,
                         "accept": "application/json",
                         "Content-Type": "application/x-www-form-urlencoded"}
        body_params = {"exSeg": exchange_segment, "prc": price, "prcTp": order_type, "prod": product, "qty": quantity,
                       "tok": instrument_token, "trnsTp": transaction_type, "trgPrc": trigger_price,
                       "brkName": broker_name, "brnchId": branch_id, "slAbsOrTks": stop_loss_type,
                       "slVal": stop_loss_value, "sqrOffAbsOrTks": square_off_type, "sqrOffVal": square_off_value,
                       "trailSL": trailing_stop_loss, "tSLTks": trailing_sl_value}
        url = "".join([self.BASEURL, self.ENDPOINTS['margin'].format(hsServerId = self.hsServerId)])
        return self._requestAPI("POST", url, data = body_params, header = header_params)

    def _requestAPI(self, method, url, body = None, data = None, header = None, timeout = 15):

        header = self.header if header == None else header

        resp = self.reqsession.request(method, url, json = body, data = data,timeout = timeout,
                                       headers = header)
        
        try: 
            data = resp.json()
            if resp.status_code == 200 : 
                return {"error" : False, "status" : True, "data" : [data], }
            
            else: 
                return {"error" : False, "status" : True, "data" : [data] }
        
        except requests.exceptions.Timeout:
            return {"status" : False, "data" : [], "error" : True, "message" : "Timeout error"}
        
        except Exception as e : 
            return {"status" : False, "data" : [], "error" : True, "message" : f"{str(e)} : {resp.text}"}
        
        return resp
        
        
        


import datetime
import traceback
from .dependencies import FyersDataSocket, SparkLib
import os 

class FyersWs():
    TOKENS = {}
    BROKER = "FYERS"

    def __init__(self, accountData, accessToken = None, dataType = "ltp", 
                 onLtp = None, 
                 onDepth = None, 
                 onError = None, 
                 onClose = None,
                 onOpen = None,
                 ) :
        self.accountData = accountData
        self.dType = "SymbolUpdate" if dataType == "ltp" else 1502 if dataType == "DepthUpdate" else None #Currently Only supports LTP Updates.
        self.onLtp = self.__onLTP if not onLtp  else onLtp
        self.onDepth = self.__onDepth if not onDepth else onDepth
        self.onError = self.__onError if not onError else onError
        self.onClose = self.__onClose if not onClose else onClose
        self.onOpen = self.__onOpen if not onOpen else onOpen
        self.run = True
        if not accessToken :
            if os.environ.get("MATICALGOS_AccessToken"):
                accessToken = os.environ["MATICALGOS_AccessToken"] 
            else: 
                raise Exception("Please generate access token.")
        self.spk = SparkLib(access_token=accessToken)
    
    def __onLTP(self, message):
        print(message)
    
    def __onDepth(self, message):
        print(message)
    
    def __onError(self, message):
        print("error : ",message)

    def __onClose(self, *args):
        print("Connection Closed")

    def __onOpen(self):
        self.Subscribe([])
        print("Connection Opened")
    
    def getTokens(self, tokens):
        data = self.spk.getBrokerTokens(tokens=tokens, broker=self.BROKER)
        if data['status'] and not data['error'] : 
            tokens = data['data']
            return tokens
        else: 
            raise Exception (data['message'])
        
    def getSymbol(self, symbol):
        return self.TOKENS[symbol]

    def tickHandler(self, message):
        try: 
            if message.get("type") in ['if', "sf"] : 
                token = self.getSymbol(message['symbol']) #self.Tokens[message['symbol']]
                msg = {"timestamp_str" : str(datetime.datetime.fromtimestamp(message['exch_feed_time'])),
                       "timestamp" : str(message['exch_feed_time']),
                       "symbol" : token, 
                       "ltp" : float(message['ltp']), 
                       "prev_day_close" : float(message['prev_close_price']), 
                       "oi" : 0, 
                       "prev_day_oi" : 0, 
                       "turnover" : int(message['vol_traded_today']) if message.get('vol_traded_today') != None else 0, 
                       "best_bid_price" : float(message['bid_price']) if message.get('bid_price') != None else 0, 
                       "best_ask_price" : float(message['ask_price']) if message.get('ask_price') != None else 0, 
                       "best_bid_qty" : int(message['bid_size']) if message.get('bid_size') != None else 0,
                       "best_ask_qty" : int(message['ask_size']) if message.get('ask_size') != None else 0, 
                       "ttq" : int(message['vol_traded_today']) if message.get('vol_traded_today') != None else 0,  
                       "token" : token}
                self.onLtp(msg)
        except Exception as e : 
            self.onError(f"Error : {e}, Traceback : {traceback.format_exc()}")

    def disconnect(self):
        self.run = False
        self.fyers.close_connection()
    
    def Subscribe(self, tokens):
        try: 
            toSubs = []
            notFound = []
            tokens = self.getTokens(tokens)
            for tk in tokens :
                wsToken = tk['wsToken']
                if wsToken != None : 
                    toSubs.append(wsToken)
                    self.TOKENS[wsToken] = tk["token"]
                else: 
                    notFound.append(tk['token'])
            
            if toSubs != [] : 
                self.fyers.subscribe(symbols=toSubs, data_type=self.dType)

            if notFound != [] : 
                self.onError(f"ERROR : Unable to find tokens for {notFound} ")
        except Exception as e: 
            self.onError(f"Error : {e}. Traceback : {traceback.format_exc()}")

    def Unsubscribe(self, tokens):
        try: 
            toUnsubs = []
            notFound = []
            tokens = self.getTokens(tokens)
            for tk in tokens :
                wsToken = tk['wsToken']
                if wsToken != None : 
                    toUnsubs.append(wsToken)
                else: 
                    notFound.append(tk['token'])

            if toUnsubs != [] : 
                self.fyers.unsubscribe(symbols=toUnsubs, data_type=self.dataType)

            if notFound != [] : 
                self.onError(f"ERROR : Unable to find tokens for {notFound} ")
                
        except: 
            self.onError("Error with Unsubscribe")

    def connect(self):
        try: 
            self.fyers = FyersDataSocket(
                access_token=self.accountData["Sessionid"],       
                log_path="",                    
                litemode=False,                  
                write_to_file=False,              
                reconnect=True,                  
                on_connect=self.onOpen,             
                on_close=self.onClose,             
                on_error=self.onError,
                on_message=self.tickHandler,
            )
            self.fyers.connect()
        except Exception as e : 
            self.onClose(f"Error : {e}, Traceback : {traceback.format_exc()}")



from .fyersHist import FyersHist
from .angelHist import angelAPI as AngelHist
from .IiflHist import IiflHist as iiflHist
from .ShoonyaHist import ShoonyaHist
from .upstoxHist import UpstoxHist
from ....sparkLib import SparkLib



AUTH_ERROR_CODE = 11001
SUBS_ERROR_CODE = 11011
UNSUBS_ERROR_CODE = 11012
RESUME_ERROR_CODE = 11031
RESUME_ERROR_CODE = 11032
MODE_ERROR_CODE = 12001
SUCCESS_CODE = 200
INVALID_CODE = -300
TOKEN_EXPIRED = -99
LIMIT_EXCEED_CODE = -99
LIMIT_EXCEED_MSG_5000 = "Please provide less than 5000 symbols"
TOKEN_EXPIRED_MSG = "Token is expired"
INVALID_TOKEN = "Please provide valid token"
SUCCESS = "ok"
ERROR = "error"
AUTH_SUCCESS = "Authentication done"
AUTH_FAIL = "Authentication failed"
SUBSCRIBE_SUCCESS = "Subscribed"
SUBSCRIBE_FAIL = "subscription failed"
UNSUBSCRIBE_SUCCESS = "Unsubscribed"
UNSUBSCRIBE_FAIL = "unsubscription failed"
LITE_MODE = "Lite Mode On"
FULL_MODE = "Full Mode On"
MODE_CHANGE_ERROR = "Mode change failed"
CHANNEL_PAUSED = "Channel Paused"
CHANNEL_RESUMED = "Channel Resumed"
CHANNEL_CHANGE_FAIL = "Mode change failed"
INVALID_SYMBOLS = "Please provide a valid symbol"
CONNECTION_CLOSED = "Connection Closed"
INDEX_DEPTH_ERROR_MESSAGE = "Index does not have market depth"
AUTH_TYPE = "cn"
SUBS_TYPE = "sub"
UNSUBS_TYPE = "unsub"
LITE_MODE_TYPE = "lit"
FULL_MODE_TYPE = "ful"
CH_PAUSE_TYPE = "cp"
CH_RESUME_TYPE = "cr"


from typing import Any, Dict, Union
from aws_lambda_powertools import Logger


class FyersLogger(Logger):
    def __init__(
        self, service: str, level: str, stack_level: int = 4, **kwargs
    ) -> None:
        """Create FyersLogger object

        Args:
            service (str): Service name. This should be the same across the code wherever the logger is initialized
            level (str): Logger level. Possible values are [INFO, DEBUG, CRITICAL, WARNING]

        Kwargs:
            stack_level (int): Stack level. This decides how many levels the stack should go up to get the line number
                            to be printed in the logging statement
        """
        super().__init__(
            service=service,
            level=level,
            location="[%(funcName)s:%(lineno)s] %(module)s",
            **kwargs
        )
        self.__stacklevel = stack_level

    def __populate_request_data(self, stack_level, **kwargs) -> Dict[str, Any]:
        """Adds additional log data to log statement

        Returns:
            Dict[str, Any]: all the keyword arguments along with extra data
        """
        kwargs["stacklevel"] = stack_level
        if "extra" not in kwargs:
            kwargs["extra"] = {}

        if "message" in kwargs["extra"]:
            kwargs["extra"]["passed_message"] = kwargs["extra"].pop("message")
        return kwargs

    def error(self, msg: Union[str, Dict[Any, Any]], *args, **kwargs) -> None:
        """Logs error statement

        Args:
            msg (Union[str, Dict[Any, Any]]): Can be str or dict object

        Kwargs:
            extra (Dict[Any, Any]): Adds this data to the log statement
        """
        stacklevel = self.__stacklevel
        while stacklevel > 0:
            try:
                kwargs = self.__populate_request_data(stacklevel, **kwargs)
                super().error(msg, *args, **kwargs)
                break
            except:
                stacklevel -= 1

    def info(self, msg: Union[str, Dict[Any, Any]], *args, **kwargs) -> None:
        """Logs info statement

        Args:
            msg (Union[str, Dict[Any, Any]]): Can be str or dict object

        Kwargs:
            extra (Dict[Any, Any]): Adds this data to the log statement
        """
        stacklevel = self.__stacklevel
        while stacklevel > 0:
            try:
                kwargs = self.__populate_request_data(stacklevel, **kwargs)
                super().info(msg, *args, **kwargs)
                break
            except:
                stacklevel -= 1

    def debug(self, msg: Union[str, Dict[Any, Any]], *args, **kwargs) -> None:
        """Logs debug statement

        Args:
            msg (Union[str, Dict[Any, Any]]): Can be str or dict object

        Kwargs:
            extra (Dict[Any, Any]): Adds this data to the log statement
        """
        stacklevel = self.__stacklevel
        while stacklevel > 0:
            try:
                kwargs = self.__populate_request_data(stacklevel, **kwargs)
                super().debug(msg, *args, **kwargs)
                break
            except:
                stacklevel -= 1

    def exception(self, msg: Union[str, Dict[Any, Any]], *args, **kwargs) -> None:
        """Logs exception statement. Should be called only from exception block

        Args:
            msg (Union[str, Dict[Any, Any]]): Can be str or dict object

        Kwargs:
            extra (Dict[Any, Any]): Adds this data to the log statement
        """
        stacklevel = self.__stacklevel
        while stacklevel > 0:
            try:
                kwargs = self.__populate_request_data(stacklevel, **kwargs)
                super().exception(msg, *args, **kwargs)
                break
            except:
                stacklevel -= 1



import zmq
import time 
import json 
import traceback
import threading
import datetime
import redis
import pandas as pd
import duckdb
import queue

class wsLib():
    def __init__(self, host = "localhost", ports = ['8542']):
        self.host = host
        self.ports = ports
        self.__connect()

    def __connect(self):
        context = zmq.Context()
        socket = context.socket(zmq.REQ)
        socket.setsockopt( zmq.LINGER,100)
        socket.setsockopt( zmq.AFFINITY,1) 
        socket.setsockopt( zmq.RCVTIMEO, 20000) # 20 Seconds timeout

        for port in self.ports:
            socket.connect ("tcp://{host}:{port}".format(host = self.host, port = port))
        return socket
    
    def connect(self, AccountName, connectionType = "ltp"):
        """
        connectionType: str, values : 'ltp' or 'depth'
        """
        msg = {"function": "startWS", "payload" : {"AccountName" : AccountName, "connectionType" : connectionType}}
        return self._request(msg)
    
    def reconnect(self, AccountName):
        msg = {"function": "reconnectWs", "payload" : {"AccountName" : AccountName}}
        return self._request(msg)
    
    def stop(self, AccountName):
        msg = {"function": "stopWs", "payload" : {"AccountName" : AccountName}}
        return self._request(msg)
    
    def subscribe(self, AccountName, tokens:list):
        msg = {"function": "subscribe", "payload" : {"AccountName" : AccountName, "tokens" : tokens}}
        return self._request(msg)
    
    def unsubscribe(self, AccountName, tokens:list):
        msg = {"function": "unsubscribe", "payload" : {"AccountName" : AccountName, "tokens" : tokens}}
        return self._request(msg)
    
    def connections(self):
        msg = {"function":"connections", "payload" : {}}
        return self._request(msg)
    
    def _request(self, msg):
        socket = self.__connect()
        try : 
            socket.send(json.dumps(msg).encode())
            response = json.loads(socket.recv().decode())
            return response
        
        except zmq.Again:
            raise Exception("Timeout")
        except zmq.ZMQError:
            raise Exception("Timeout")
        except Exception as e : 
            print(traceback.print_exc())
            raise Exception(e)

import ast

class dataStream():
    _run = True
    tokens = []
    __allTicks = False
    __allMin = False
    timeout = 2.5
    MAX_RECONNECT = 10
    RECONNECT_NO = 0
    isReconnecting = False
    __thRunner = {}
    
    def __init__(self, host = "localhost", port = 8543):
        self.host = host
        self.port = port
        self.tickStream = self.__tickStream
        self.minStream = self.__minStream
        self.lastUpdate = time.time()
        self.lock = threading.Lock()
        self.ReconQueue = queue.Queue()
        self.is_recon = False
        self.maxReconAlert = False
        
    def __tickStream(self, message):
        print(message)
        
    def __minStream(self, message):
        print(message)
    
    # def __connect(self):
    #     try: 
    #         while self._run : 
    #             try : 
    #                 data = self.socket.recv(zmq.NOBLOCK) #flags = zmq.NOBLOCKS
    #                 # print(data)
    #                 # token, message = data.decode().split(" _&_ ")
    #                 # if token in self.tokens : 
    #                 #     dt = json.loads(message)
    #                 #     self.__updateHandler(message = dt, isTick = True if "tick" in token else False)
                    
    #                 # msgpack
    #                 # message = msgpack.unpackb(data)
    #                 # token = message['token']
    #                 # if token in self.tokens : 
    #                 #     self.__updateHandler(message = message, isTick = True if message['type'] == "tick" else False)

    #                 data = ast.literal_eval(data.decode())
    #                 for d in data : 
    #                     token, message = d.split(" _&_ ")
    #                     if token in self.tokens or ("tick" in token and self.__allTicks) or ("min" in token and self.__allMin) : 
    #                         dt = json.loads(message)
    #                         self.__updateHandler(message = dt, isTick = True if "tick" in token else False)
                
    #             except zmq.Again : 
    #                 time.sleep(0.01)
    #                 pass
                
    #             except zmq.ZMQError as e : 
    #                 traceback.print_exc()
    #                 print(e)
    #                 self._run = False

    #             except Exception as e:
    #                 traceback.print_exc()
    #                 print(f"Unexpected error: {e}")
    #                 self._run = False
                    
    #     finally: pass
    
    def __connect(self, key = None):
        try: 
            self.context = zmq.Context().instance()
            self.socket = self.context.socket(zmq.SUB)
            self.socket.setsockopt(zmq.LINGER, 100)
            self.socket.setsockopt_string(zmq.SUBSCRIBE, "")
            self.socket.connect("tcp://{host}:{port}".format(host = self.host, port = self.port))
            self.poller = zmq.Poller()
            self.poller.register(self.socket, zmq.POLLIN)
            key = int(time.time())
            self.__thRunner[key] = True
            print(self.__thRunner, key)
            while  self.__thRunner[key] : 
                try : 
                    events = dict(self.poller.poll(timeout=100))
                    if events : 
                        self.lastUpdate = time.time()
                        if self.socket in events : 
                            data = self.socket.recv(zmq.NOBLOCK) #flags = zmq.NOBLOCKS
                            data = ast.literal_eval(data.decode())
                            for d in data : 
                                token, message = d.split(" _&_ ")
                                if token in self.tokens or ("tick" in token and self.__allTicks) or ("min" in token and self.__allMin) : 
                                    dt = json.loads(message)
                                    self.__updateHandler(message = dt, isTick = True if "tick" in token else False)
                    else: 
                        if time.time() - self.lastUpdate > self.timeout and self.tokens != [] and not self.isReconnecting : 
                            if self.RECONNECT_NO < self.MAX_RECONNECT : 
                                print("RECON", self.RECONNECT_NO)
                                self.RECONNECT_NO += 1 
                                with self.lock : 
                                    self.__thRunner[key] = False
                                    self.isReconnecting = True
                            elif not self.maxReconAlert: 
                                print(f"We've reached max reconnections of {self.MAX_RECONNECT}, There may be some issue with the data feed.") 
                                self.maxReconAlert = True
                except zmq.ZMQError as e : 
                    traceback.print_exc()
                    print(e)
                    self.__thRunner[key] = False

                except Exception as e:
                    traceback.print_exc()
                    print(f"Unexpected error: {e}")
                    self.__thRunner[key] = False
                    
            if not self.__thRunner[key] : 
                if self.isReconnecting:
                    self.ReconQueue.put(key)
        
        finally:
            pass 
    
    def reconQueue(self):
        while self._run:
            try: 
                data = self.ReconQueue.get(timeout=1)
                self.stop(restart=True, key=data) 
                time.sleep(5)
                self.connect()
                with self.lock:
                    self.isReconnecting = False
                    
                time.sleep(0.1)
            except queue.Empty:
                pass
    
    def stop(self, restart = False, key = None):
        if not restart: 
            self._run = False
            with self.lock:
                for key in self.__thRunner.keys():
                    self.__thRunner[key] = False
            time.sleep(1)
        with self.lock:
            if key != None : 
                self.__thRunner[key] = False  
        self.poller.unregister(self.socket)
        self.socket.close()
        self.context.term()

    def connect(self):
        if not self.is_recon:
            threading.Thread(target=self.reconQueue).start()
            self.is_recon = True
        th = threading.Thread(target = self.__connect, daemon=True).start()

    def reconnect(self):
        self.RECONNECT_NO+=1 
        self.stop()
        with self.lock:
            while any(self.__thRunner.values()):
                time.sleep(0.1)
        self.connect()
        with self.lock:
            self.isReconnecting = False

    def subscribeAll(self, ticks = True):
        if ticks: 
            self.__allTicks = True
        else: 
            self.__allMin = True
            self.tokens.append("min:*")

    def subscribe(self, tokens:list, isTick = True):
        for t in tokens : 
            tok = f"min:{t}" if not isTick else f"tick:{t}"
            self.socket.setsockopt_string(zmq.SUBSCRIBE, tok)
            self.tokens.append(tok)
    
    def unsubscribe(self, tokens:list, isTick = True):
        for t in tokens : 
            tok = f"min:{t}" if not isTick else f"tick:{t}"
            self.socket.setsockopt_string(zmq.UNSUBSCRIBE, tok)
            self.tokens.remove(tok)
           
    def __updateHandler(self, message, isTick = True):
        if isTick : 
            self.tickStream(message)
        else: 
            self.minStream(message)

class orderwsStream():
    _run = True
    
    def __init__(self, host = "localhost", port = 8544):
        self.host = host
        self.port = port
        self.__ZMQcontext = zmq.Context()
        self.socket = self.__ZMQcontext.socket(zmq.SUB)
        self.socket.setsockopt_string(zmq.SUBSCRIBE, "")
        self.updates = self.__wsUpdates
    
    def __wsUpdates(self, message):
        print(message)
    
    def __connect(self):
        self.socket.connect("tcp://{host}:{port}".format(host = self.host, port = self.port))
        while self._run : 
            try : 
                data = self.socket.recv() #flags = zmq.NOBLOCKS
                data = json.loads(data.decode())
                self.updates(data)
            except : 
                traceback.print_exc()
            time.sleep(0.01)
    
    def stop(self):
        self._run = False
        self.socket.close()
        self.__ZMQcontext.term()
        self.__ZMQcontext.destroy()
    
    def connect(self):
        threading.Thread(target = self.__connect).start()

class redisData():
    defaultStart = datetime.datetime.now().replace(hour=9,minute=15,second=0,microsecond=0)
    defaultEnd = datetime.datetime.now().replace(hour=15,minute=30,second=0,microsecond=0)
    def __init__(self, host = "localhost", port = 6379, password=None):
        self.host = host
        self.port = port
        self.password = password
        self.__pool()
    
    def __pool(self):
        self.pool = redis.ConnectionPool(host = self.host, port = self.port)
    
    def __connect(self):
        return redis.Redis(connection_pool=self.pool,password=self.password)
     
    def getMin(self, token, starttime = "", endtime = "", timeframe = 1, origin = datetime.time(9,15)):
        starttime = self.defaultStart if starttime == "" else starttime
        endtime = self.defaultEnd if endtime == "" else endtime

        start_ts = int((starttime).timestamp())
        end_ts = int((endtime).timestamp())
        r = self.__connect()
        data = r.zrangebyscore(f"min:{token}", start_ts, end_ts)
        data = [json.loads(i.decode()) for i in data]
        if data == [] : 
            return pd.DataFrame()
        df = pd.DataFrame(data)
        df = df[['ts','tk','o','h','l','c','v','oi']]
        df = df.rename(columns={"ts":"datetime",'tk':'token',"o":"open","h":'high','l':"low","c":"close","v":"volume"})
        df['datetime'] = pd.to_datetime(df['datetime'], format="%Y-%m-%d %H:%M:%S")
        df = df.set_index('datetime')
        initVal = df['volume'].iloc[0]
        df[['volume']] = df[['volume']].fillna(0)
        df["volume"] = df['volume'] - df["volume"].shift(1, fill_value = 0)
        df.loc[df.index[0], 'volume'] = initVal

        if timeframe != 1 : 
            df = df.resample(f'{str(timeframe)}min', origin = datetime.datetime.combine(df.iloc[0].name.date(), origin)).agg({"open" : "first", 
                                                          "token" : "last",
                                                          "high" : "max", 
                                                          "low" : "min", 
                                                          "close" : "last", 
                                                          "volume" : "sum", 
                                                          "oi" : "last"
                                                          }).dropna()
        return df
    
    def getMinCandle(self, token, dt):
        r = self.__connect()
        ts = int(dt.timestamp())
        data = r.zrangebyscore(f"min:{token}", ts, ts)
        if data == [] : 
            return None
        return json.loads(data[0].decode())


class histDB():
    def __init__(self, filename, path:str="./"):
        self.filename=filename
        self.path=path
        self.startTime= datetime.datetime.now().replace(hour=0,minute=0,second=0,microsecond=0)-datetime.timedelta(days=1)
        self.endTime = datetime.datetime.now().replace(hour=23,minute=0,second=0,microsecond=0)-datetime.timedelta(days=1)
        self.columns = "token,datetime,open,high,low,close,volume,oi"
    def __connect(self):
        return duckdb.connect(r"{path}/{filename}.db".format(path=self.path,filename=self.filename), read_only=True)

    def fetchData(self, token, startTime:datetime=None, endTime:datetime=None, timeframe:int=1, origin = datetime.time(9,15)):
        conn = self.__connect()
        startTime = self.startTime if startTime == None else startTime
        endTime = self.endTime if endTime == None else endTime
        df =  conn.execute(f"""SELECT {self.columns} FROM dbtable WHERE token='{token}' AND datetime >= '{str(startTime)}' AND datetime <= '{str(endTime)}' ORDER BY datetime ASC""").df()
        df['datetime'] = pd.to_datetime(df['datetime'], format="%Y-%m-%d %H:%M:%S")
        df = df.set_index('datetime')
        if timeframe != 1 : 
            df = df.resample(f'{str(timeframe)}min', origin = datetime.datetime.combine(df.iloc[0].name.date(), origin)).agg({"open" : "first", 
                                                          "high" : "max", 
                                                          "low" : "min", 
                                                          "close" : "last", 
                                                          "volume" : "sum", 
                                                          "token" : "last",
                                                          "oi" : "last"
                                                          }).dropna()
        return df

class tradingHistData():
    def __init__(self, redisconf = {"host":"localhost","port":"6379","password":None},
                       historical = False, filename:str=None, path:str="./"):
        self.historical=False
        if historical:
            self.hist = histDB(filename=filename,path=path)
            self.historical = True
        self.redisData = redisData(**redisconf)
        self.startTime= datetime.datetime.now().replace(hour=9,minute=15,second=0,microsecond=0)
        self.endTime = datetime.datetime.now().replace(hour=15,minute=30,second=0,microsecond=0)

    def _resample(self, df, timeframe, origin = datetime.time(9,15)):
        df = df.resample(f'{str(timeframe)}min', origin = datetime.datetime.combine(df.iloc[0].name.date(), origin)).agg({"open" : "first", 
                                                          "high" : "max", 
                                                          "low" : "min", 
                                                          "close" : "last", 
                                                          "volume" : "sum", 
                                                          "token" : "last",
                                                          "oi" : "last"
                                                          }).dropna()
        return df 
    
    def fetchData(self, token:str,startTime:datetime=None,endTime:datetime=None, timeframe:int=1):
        startTime = self.startTime if startTime == None else startTime
        endTime = self.endTime if endTime == None else endTime
        if startTime < self.startTime and self.historical:
            data = self.hist.fetchData(token=token,startTime=startTime,
                                       endTime=endTime, timeframe=timeframe)
        else: 
            data = pd.DataFrame()
            
        if endTime > self.startTime:
            dataTd = self.redisData.getMin(token=token, starttime=startTime,endtime=endTime,timeframe=timeframe)
        else: 
            dataTd = pd.DataFrame()

        if not data.empty and not dataTd.empty : 
            master = pd.concat([data,dataTd])
            return master
        
        if not data.empty: return data
        if not dataTd.empty: return dataTd





import datetime
import os 
from .dependencies import SparkLib, FyersHist, AngelHist, iiflHist, UpstoxHist, ShoonyaHist 

BROKERS = {
    "FYERS" : FyersHist,
    "ANGELONE" : AngelHist,
    "SHOONYA" : ShoonyaHist,
    # "KOTAKNEO" : "", #Historical data not available for kotak.
    "UPSTOX" : UpstoxHist,
    "IIFLXTSDATA" : iiflHist
}

class HistoricalData():
    accounts = {}
    startTime = datetime.datetime.now().replace(hour=9,minute=15, second=0, microsecond=0)
    endTime = datetime.datetime.now().replace(hour=15,minute=30, second=0, microsecond=0)
    __today = str(datetime.datetime.now().date())

    def __init__(self, accessToken:str = None):
        if not accessToken :
            if os.environ.get("MATICALGOS_AccessToken"):
                accessToken = os.environ["MATICALGOS_AccessToken"] 
            else: 
                raise Exception("Please generate access token.")
        self.spk = SparkLib(access_token=accessToken)

    def initiateAccount(self, accountName):
        data = self.spk.getOneAccount(accountName=accountName)
        if data['status'] and not data['error']:
            accountData = data['data'][0]
            if accountData['LastLogin'] == self.__today : 
                broker=accountData['Broker']
                sessionData= accountData['Sessionid']
                self.accounts[accountName] = {"sessionid" : sessionData,
                                              "broker" : broker,
                                              "accountName" : accountName,
                                              "accountData" : accountData,
                                              "__function" : BROKERS[broker](accountData)}
            else: 
                raise Exception(f"Generate Code for account Name : {accountName}")
        else: 
            raise Exception(data['message'])

    def _info(self, accountName):
        if accountName not in self.accounts.keys():
            raise Exception(f"Account Name {accountName} not found.")
        return self.accounts[accountName]["__function"].info()

    def getToken(self, token:list, broker):
        broker = "IIFLXTS" if broker == "IIFLXTSDATA" else broker
        tokendata = self.spk.getBrokerTokens(token, broker=broker)
        if tokendata['status'] and not tokendata['error'] : 
            tokenName = tokendata['data']
            if tokenName != [] : 
                return tokenName
            else : 
                raise Exception(f'broker token not found for {token}')
        else: 
            raise Exception(tokendata['message'])

    def getHistorical(self, accountName:str, 
                      token:str,
                      startTime:datetime = startTime, 
                      endTime:datetime=endTime, 
                      interval = "1", reInitialize = False):
        
        if accountName not in self.accounts.keys() or reInitialize :
            self.initiateAccount(accountName)
        accountData = self.accounts[accountName]
        token = self.getToken([token], broker=accountData['broker'])
        if token[0]['histToken'] != None : 
            data = accountData['__function'].getHistorical(token = token[0]['histToken'], 
                                                        startTime = startTime, 
                                                        endTime = endTime, 
                                                        interval = interval)
            return data
        else: 
            raise Exception("Token not found.")
    
    def getQuotes(self, accountName : str, tokens : list, reInitialize = False):
        if accountName not in self.accounts.keys() or reInitialize :
            self.initiateAccount(accountName)
        accountData = self.accounts[accountName]
        tokens = self.getToken(tokens, broker=accountData['broker'])
        # tokens = [i['histToken'] for i in tokens]
        data = accountData['__function'].get_quotes(tokens)
        return data




